#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include "uintset.h"
#include "aliquot_sum.h"

void find_sequence(int fd, unsigned int n) {
   unsigned int last = n;
   uintset members;
   if (!uintset_init(members)) {
      perror("uintset_init"); exit(1);
   }
   for(;;) {
      uintset_add(members, last);
      unsigned int sum = aliquot_sum(last);
      if (sum == n) {
	 /* found something, pass it to fd */
	 if (write(fd, &n, sizeof n) < sizeof n) {
	    perror("write error"); exit(1);
	 }
	 break;
      }
      if (sum < n || uintset_ismember(members, sum)) {
	 /* already processed or out of range */
	 break;
      }
      last = sum;
   }
   uintset_clear(members);
}

void print_sequence(unsigned int n) {
   printf("%u", n);
   unsigned int sum = n;
   while ((sum = aliquot_sum(sum)) && sum != n) {
      printf(", %u", sum);
   }
   printf("\n");
}

void process_range(int fd, unsigned int min, unsigned int max) {
   for (int n = min; n <= max; ++n) {
      find_sequence(fd, n);
   }
}

int main(int argc, char** argv) {
   char* cmdname = *argv++; --argc;
   char* usage = "Usage: %s nofprocesses a b\n";
   if (argc != 3) {
      fprintf(stderr, usage, cmdname); exit(1);
   }
   unsigned int nofprocesses = atoi(*argv++); --argc;
   unsigned int a = atoi(*argv++); --argc;
   unsigned int b = atoi(*argv++); --argc;
   if (nofprocesses <= 0 || a <= 0 || b <= 0 || a > b || b-a+1 < nofprocesses) {
      fprintf(stderr, usage, cmdname); exit(1);
   }

   unsigned int size = b - a + 1;
   unsigned int remainder = size % nofprocesses;
   unsigned int start = a;

   pid_t pid[nofprocesses];
   int fd[nofprocesses];

   for (int i = 0; i < nofprocesses; ++i) {
      unsigned int len = size / nofprocesses;
      if (remainder < i) len += 1;
      int fds[2];
      if (pipe(fds) < 0) {
	 perror("pipe"); exit(1);
      }
      fd[i] = fds[0];
      pid[i] = fork();
      if (pid[i] < 0) {
	 perror("fork"); exit(1);
      }
      if (pid[i] == 0) {
	 /* close reading side */
	 close(fds[0]);
	 /* close all other pipes */
	 for (int j = 0; j < i; ++j) {
	    close(fd[j]);
	 }
	 process_range(fds[1], start, start + len - 1);
	 exit(0);
      }
      /* close writing side */
      close(fds[1]);
      start += len;
   }
   for (int i = 0; i < nofprocesses; ++i) {
      unsigned int n;
      while (read(fd[i], &n, sizeof(unsigned int)) == sizeof(unsigned int)) {
	 print_sequence(n);
      }
      close(fd[i]);
      int stat;
      if (waitpid(pid[i], &stat, 0) < 0) {
	 perror("wait");
      } else if (!WIFEXITED(stat) || WEXITSTATUS(stat)) {
	 fprintf(stderr, "%s: process %d terminated abnormally\n",
	    cmdname, (int)pid[i]);
      }
   }
}
