#include <stdlib.h>
#include <stralloc.h>
#include <string.h>
#include <afblib/inbuf.h>
#include <afblib/inbuf_scan.h>
#include <afblib/pconnect.h>
#include <afblib/strlist.h>
#include <sys/wait.h>
#include "weather.h"

static char hexdigit(unsigned int val) {
   if (val < 10) {
      return '0' + val;
   } else {
      return 'A' + (val - 10);
   }
}

static bool stralloc_cat_encoded_string(stralloc* encoded,
      char* s, size_t len) {
   for (unsigned int i = 0; i < len; ++i) {
      char ch = s[i];
      if (ch >= '0' && ch <= '9' || ch >= 'A' && ch <= 'Z' ||
	    ch >= 'a' && ch <= 'z') {
	 if (!stralloc_readyplus(encoded, 1)) return false;
	 encoded->s[encoded->len++] = ch;
      } else {
	 if (!stralloc_readyplus(encoded, 3)) return false;
	 encoded->s[encoded->len++] = '%';
	 encoded->s[encoded->len++] = hexdigit(ch / 16);
	 encoded->s[encoded->len++] = hexdigit(ch % 16);
      }
   }
}

static bool stralloc_cat_encoded(stralloc* encoded, stralloc* s) {
   stralloc_cat_encoded_string(encoded, s->s, s->len);
}

static bool stralloc_cat_encoded_s(stralloc* encoded, char* s) {
   stralloc_cat_encoded_string(encoded, s, strlen(s));
}

bool init_weather(Weather* weather, char* location) {
   return stralloc_copys(&weather->location, location);
}

bool update_weather(Weather* weather) {
   strlist argv = {0};
   strlist_push(&argv, "wget");
   strlist_push(&argv, "-O-");
   strlist_push(&argv, "-q");
   strlist_push(&argv, "--no-check-certificate");
   stralloc url = {0};
   stralloc_copys(&url, "https://query.yahooapis.com/v1/public/yql?q=");
   stralloc_cat_encoded_s(&url,
      "select item.condition from weather.forecast "
      "where u='c' and woeid in ("
	 "select woeid from geo.places(1) where text=\"");
	 stralloc_cat_encoded(&url, &weather->location);
   stralloc_cat_encoded_s(&url,
	 "\""
      ")");
   stralloc_cats(&url, "&format=json");
   if (!stralloc_0(&url)) return false;
   strlist_push(&argv, url.s);
   if (!strlist_push0(&argv)) return false;

   pipe_end pipe;
   bool ok = pconnect(argv.list[0], argv.list, PIPE_READ, &pipe);
   if (ok) {
      inbuf in = {pipe.fd};
      stralloc temp = {0};
      if (inbuf_scan(&in, ".*?"
	       "\"temp\":\\s*\"(\\d+)\""
	       ".*?"
	       "\"text\":\\s*\"(.*?)\"",
	    &temp, &weather->text) == 2) {
	 stralloc_0(&temp);
	 weather->temp = atoi(temp.s);
      } else {
	 ok = false;
      }
      stralloc_free(&temp);
      phangup(&pipe);
      ok = WIFEXITED(pipe.wstat) && WEXITSTATUS(pipe.wstat) == 0 && ok;
   }
   stralloc_free(&url);
   strlist_free(&argv);
   return ok;
}

void free_weather(Weather* weather) {
   stralloc_free(&weather->location);
}
